unit HrastUnit;

// version: 15
// Copyright  2014 by HrastProgrammer. All rights reserved.
// This unit contains WREN synthesizer modules developed by HrastProgrammer.
// Can be distributed only as part of the complete WREN source package.
//
// version  8: Blue Hell : changed module registration procedure.
// version  9: Blue Hell : layout changes.
// version 10: Blue Hell : some code changes.
// version 11: Blue Hell : added constructors for the Fast/Slow selection.
// version 12: Blue Hell : removal of outcommented code and object renaming.
// version 13: Blue Hell : added un-attenuated FM to HrastSVF.
// version 14: Blue Hell : changed the lights gatherer.
// version 15: Blue Hell : added some normalization to the filter.
// version 16: Blue Hell : renamed some stuff to make things cinsistent with other modules - theis will break some patches

interface

uses

  KnobsUtils, KnobsConversions, Module_Defs, Module_Implementations;

type

  TModHrastOsc = class( TMod)
  strict private
  const
    i_frequency =  0;
    i_morph     =  1;
    i_shape     =  2;
    i_bandlimit =  3;
    i_pw        =  4;
    i_pwmlevel  =  5;
    i_pwm       =  6;
    i_fmlevel   =  7;
    i_fm        =  8;
    i_pmlevel   =  9;
    i_phase     = 10;
    i_vintage   = 11;
    i_sync      = 12;
    i_level     = 13;
    i_mute      = 14;
    i_cents     = 15;

    o_out       =  0;
  private
    FZ        : TSignal;
    FPhase    : TSignal;
    FPrevSync : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                       override;
    procedure   SetDefaults;                                                              override;
    procedure   DoTick;                                                                   override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);    override;
    function    InputName ( anIndex: Integer): string;                                    override;
    function    OutputName( anIndex: Integer): string;                                    override;
  end;

  TModHrastSVF = class( TMod)
  strict private
  const
    i_in         =  0;
    i_frequency  =  1;
    i_q          =  2;
    i_fm         =  3;
    i_fmlevel    =  4;
    i_qm         =  5;
    i_qmlevel    =  6;
    i_inlevel    =  7;
    i_distortion =  8;
    i_lowpass    =  9;
    i_bandpass   = 10;
    i_highpass   = 11;
    i_level12    = 12;
    i_level24    = 13;
    i_freq       = 14;

    o_lp         =  0;
    o_bp         =  1;
    o_hp         =  2;
    o_br         =  3;
    o_12         =  4;
    o_24         =  5;
  private
    FBand12 : TSignal;
    FLow12  : TSignal;
    FBand24 : TSignal;
    FLow24  : TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string);                       override;
    procedure   SetDefaults;                                                              override;
    procedure   DoTick;                                                                   override;
    function    InputName ( anIndex: Integer): string;                                    override;
    function    OutputName( anIndex: Integer): string;                                    override;
  end;


  procedure CreateTables; // BLue Hell : renamed inintialize to CreateTables and made it public; table re-init is needed
                          //             after changing the system rate



implementation



uses

  Globals, Math, SysUtils;

const

  TableSize = 8192;

var

  Pow2Table : array [ 0 .. TableSize + 1] of TSignal;
  SineTable : array [ 0 .. TableSize + 1] of TSignal;


  function Pow2Freq( Freq: TSignal): TSignal;
  var
    n: Integer;
    f: TSignal;
  begin
    f := Abs( Freq) * TableSize;
    if f > TableSize - 1
    then f := TableSize - 1;
    n := Trunc( f);
    f := f - n;
    Result := Pow2Table[ n];
    Result := Result + f * ( Pow2Table[ n + 1] - Result);
  end;

  function Sine( Freq: TSignal): TSignal;
  var
    n: Integer;
    f: TSignal;
  begin
    f := Abs( Freq) * TableSize;
    if f > TableSize - 1
    then f := TableSize - 1;
    n := Trunc( f);
    f := f - n;
    Result := SineTable[ n];
    Result := Result + f * ( SineTable[ n + 1] - Result);
  end;


{ ========
  TModHrastOsc = class( TMod)
  private
    FZ        : TSignal;
    FPhase    : TSignal;
    FPrevSync : Boolean;
  public
}

    constructor TModHrastOsc.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModHrastOsc.SetDefaults;
    begin
      FInputs[ i_fm   ] := 0;
      FInputs[ i_phase] := 0;
      FInputs[ i_pwm  ] := 0;
      FDezipperMap := [
        i_frequency,
        i_morph,
        i_shape,
        i_bandlimit,
        i_pw,
        i_pwmlevel,
        i_fmlevel,
        i_pmlevel,
        i_level,
        i_mute,
        i_cents
      ];
    end;

    procedure TModHrastOsc.DoTick;
    var
      DeltaPhase: TSignal;
    var
      n   : Integer;
      t   : TSignal;
      y   : TSignal;
      ypw : TSignal;
    begin
      if not FPrevSync and SignalToLogic( FInputs[i_sync])
      then FPhase := 0;

      FPrevSync  := SignalToLogic( FInputs[ i_sync]);
      DeltaPhase := Pow2Freq( FInputs[ i_frequency] + FInputs[ i_cents] + FInputs[ i_fmlevel] * FInputs[ i_fm]) * System_Rate_Rec;

      t := FPhase + FInputs[ i_pmlevel] * FInputs[ i_phase];
      n := Trunc( t);
      t := t - n;
      if t < 0
      then t := t + 1;

      FPhase := FPhase + DeltaPhase;

      n      := Trunc( FPhase);
      FPhase := FPhase - n;
      if FPhase < 0
      then FPhase := FPhase + 1;

      y := 2 * t - 1 - PolyBlep( t, FInputs[ i_bandlimit], DeltaPhase);

      t := t + FInputs[ i_pw] + FInputs[ i_pwmlevel] * FInputs[ i_pwm];

      n := Trunc( t);
      t := t - n;
      if t < 0
      then t := t + 1;

      ypw := 2 * t - 1 - PolyBlep( t, FInputs[ i_bandlimit], DeltaPhase);

      t := FInputs[ i_shape];

      if FInputs[ i_vintage] <> 0
      then begin                            // Vintage
        if t > 0.999
        then t := 0.999;
        y := ( y + 1) * 0.5;
        y := y * y;
        y := 2 * y - 1;

        ypw := ( ypw + 1) * 0.5;
        ypw := ypw * ypw;
        ypw := 2 * ypw - 1;
      end;
      FZ := y - ypw * FInputs[ i_morph] + t * FZ;

      if t > 0.99
      then t := 0.99;

      y := FZ * ( t - 1);
      FOutputs[ o_out] := y * FInputs[ i_level] * SignalToMute( FInputs[ i_mute]);
    end;

    procedure TModHrastOsc.GatherLights( const aPrefix: string; const aCallback: TLightsHandler);
    begin
      aCallback( Self, MakeInfo( aPrefix, Name, 'syncinput', FInputs[ i_sync]));
    end;

    function TModHrastOsc.InputName( anIndex: Integer): string;
    begin
      case anIndex of
        i_frequency : Result := 'frequency';
        i_morph     : Result := 'morph';
        i_shape     : Result := 'shape';
        i_bandlimit : Result := 'bandlimit';
        i_pw        : Result := 'pw';
        i_pwmlevel  : Result := 'pwmlevel';
        i_pwm       : Result := 'pwm';
        i_fmlevel   : Result := 'fmlevel';
        i_fm        : Result := 'fm';
        i_pmlevel   : Result := 'pmlevel';
        i_phase     : Result := 'phase';
        i_vintage   : Result := 'vintage';
        i_sync      : Result := 'sync';
        i_level     : Result := 'level';
        i_mute      : Result := 'mute';
        i_cents     : Result := 'cents';
        else          Result := NO_INPUT;
      end
    end;

    function TModHrastOsc.OutputName( anIndex: Integer): string;
    begin
      case anIndex of
        o_out : Result := 'out';
        else    Result := NO_OUTPUT;
      end
    end;


{ ========
  TModHrastSVF = class( TMod)
  private
    FBand12 : TSignal;
    FLow12  : TSignal;
    FBand24 : TSignal;
    FLow24  : TSignal;
  public
}

    constructor TModHrastSVF.Create( aParent: TSynthPatch; const aName: string); // override;
    begin
      inherited;
      FIsFast := True;
      FIsSlow := False;
    end;

    procedure   TModHrastSVF.SetDefaults;
    begin
      FInputs[ i_in  ] := 0;
      FInputs[ i_fm  ] := 0;
      FInputs[ i_qm  ] := 0;
      FInputs[ i_freq] := 0;
      FDezipperMap:=[
        i_frequency,
        i_q,
        i_fmlevel,
        i_qmlevel,
        i_inlevel,
        i_distortion,
        i_lowpass,
        i_bandpass,
        i_highpass,
        i_level12,
        i_level24
      ]
    end;

    procedure TModHrastSVF.DoTick;
    var
      Oversample : Integer;
      f          : TSignal;
      fb         : TSignal;
      y          : TSignal;
      High12     : TSignal;
      High24     : TSignal;
      Output12   : TSignal;
      Output24   : TSignal;
    begin
      f  := Pow2Freq( FInputs[ i_frequency] + FInputs[ i_freq] + FInputs[ i_fmlevel] * FInputs[ i_fm]) / System_Rate_2;
      f  := 2 * Sine( f);
      fb := Clip( 1 - ( FInputs[ i_q] + FInputs[ i_qmlevel] * FInputs[ i_qm]), 0, 1);
      y  := FInputs[ i_in] * FInputs[ i_inlevel];
      y  := y / ( Abs( y) + ( 1 - Clip( FInputs[ i_distortion], 0, 0.999)));

      for Oversample := 1 to 2
      do begin
        High12   := Normalize( y - fb * FBand12 - FLow12);
        FBand12  := Normalize( f * High12  + FBand12);
        FLow12   := Normalize( f * FBand12 + FLow12);
        FLow12   := Normalize( FLow12 * 1.0001 / ( 0.0001 * Abs( FLow12) + 1));
        Output12 := Normalize( FLow12 * FInputs[ i_lowpass] + FBand12 * FInputs[ i_bandpass] + High12 * FInputs[ i_highpass]);

        High24   := Normalize( Output12 - fb * FBand24 - FLow24);
        FBand24  := Normalize( f * High24  + FBand24);
        FLow24   := Normalize( f * FBand24 + FLow24);
        FLow24   := Normalize( FLow24 * 1.0001 / ( 0.01 * Abs( FLow24) + 1));
      end;

      Output24 := FLow24 * FInputs[ i_lowpass] + FBand24 * FInputs[ i_bandpass] + High24 * FInputs[ i_highpass];
      FOutputs[ o_lp] := FLow12;
      FOutputs[ o_bp] := FBand12;
      FOutputs[ o_hp] := High12;
      FOutputs[ o_br] := FLow12 + High12;
      FOutputs[ o_12] := Output12 * FInputs[ i_level12];
      FOutputs[ o_24] := Output24 * FInputs[ i_level24];
    end;

    function TModHrastSVF.InputName;
    begin
      case anIndex of
        i_in          : Result := 'in';
        i_frequency   : Result := 'frequency';
        i_q           : Result := 'q';
        i_fm          : Result := 'fm';
        i_fmlevel     : Result := 'fmlevel';
        i_qm          : Result := 'qm';
        i_qmlevel     : Result := 'qmlevel';
        i_inlevel     : Result := 'inlevel';
        i_distortion  : Result := 'distortion';
        i_lowpass     : Result := 'lowpass';
        i_bandpass    : Result := 'bandpass';
        i_highpass    : Result := 'highpass';
        i_level12     : Result := 'level12';
        i_level24     : Result := 'level24';
        i_freq        : Result := 'freq';
        else            Result := NO_INPUT;
      end
    end;

    function TModHrastSVF.OutputName;
    begin
      case anIndex of
        o_lp : Result := 'lp';
        o_bp : Result := 'bp';
        o_hp : Result := 'hp';
        o_br : Result := 'br';
        o_12 : Result := '12';
        o_24 : Result := '24';
        else   Result := NO_OUTPUT;
      end
    end;



  procedure CreateTables;
  var
    i: Integer;
    x: TSignal;
  begin
    for i := 0 to TableSize
    do Pow2Table[ i] := UnitsToFrequency( i / TableSize);
    Pow2Table[ TableSize + 1] := Pow2Table[ TableSize];

    x := Pi / TableSize;
    for i := 0 to TableSize - 1
    do SineTable[ i] := Sin( x * i);
    SineTable[ TableSize div 2] := 0;
    SineTable[ TableSize      ] := SineTable[ 0];
    SineTable[ TableSize + 1  ] := SineTable[ 1];
  end;

initialization

  CreateTables;

end.
