{
  Copyright 2000 - blue hell / Jan Punter

  this file is part of the friemels program.
}



unit KnobsWorms;

interface

uses

  WinApi.Windows, System.Math, System.SysUtils, System.Classes, Vcl.Graphics, Vcl.Controls, Vcl.StdCtrls, Vcl.ExtCtrls,
  System.Types, WinApi.Messages;

type


  EKnobsWorm        = class( Exception );
  EKnobsInvalidFile = class( EKnobsWorm);
  EKnobsBounds      = class( EKnobsWorm);

  TKnobsWormElt    = Byte;
  TKnobsWormData   = array of TKnobsWormElt;
  TKnobsPercentage = 0 .. 100;

  TKnobsGene = class;

  TKnobsOnGetGeneName = procedure( const aSender: TObject; const aGene: TKnobsGene; var aName: string) of object;
  TKnobsOnWormChanged = procedure( const aSender: TObject; const aGene: TKnobsGene)                    of object;

  TKnobsWorm = class( TGraphicControl)
  private
    FGene                 : TKnobsGene;
    FAudible              : Boolean;
    FFrozen               : Boolean;
    FOnAudible            : TNotifyEvent;
    FOnFrozen             : TNotifyEvent;
    FClearance            : Byte;
    FBackColor            : TColor;
    FBorderColorNormal    : TColor;
    FBorderColorAudible   : TColor;
    FBorderColorFrozen    : TColor;
    FBorderColorBoth      : TColor;
    FPenColor             : TColor;
    FCrossOverProbability : TKnobsPercentage;
    FMutationProbability  : TKnobsPercentage;
    FMutationRange        : TKnobsWormElt;
    FMorphPercentage      : TKnobsPercentage;
    FOnWormChanged        : TKnobsOnWormChanged;
    FDrawLines            : Boolean;
  private
    procedure   MutePeers;
    procedure   UnActivatePeers;
    function    GetData( anIndex: Integer): TKnobsWormElt;
    function    GetSize              : Integer;
    procedure   SetSize              ( aValue: Integer);
    procedure   SetAudible           ( aValue: Boolean);
    procedure   SetFrozen            ( aValue: Boolean);
    procedure   SetClearance         ( aValue: Byte   );
    procedure   SetBackColor         ( aValue: TColor );
    procedure   SetBorderColorNormal ( aValue: TColor );
    procedure   SetBorderColorAudible( aValue: TColor );
    procedure   SetBorderColorFrozen ( aValue: TColor );
    procedure   SetBorderColorBoth   ( aValue: TColor );
    procedure   SetPenColor          ( aValue: TColor );
    procedure   SetDrawLines         ( aValue: Boolean);
  protected
    procedure   Paint;                                                 override;
    procedure   Do_OnDragDrop ( Sender, Source: TObject; X, Y: Integer);
    procedure   Do_OnDragOver ( Sender, Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean);
    procedure   Do_OnEndDrag  ( Sender, Target: TObject; X, Y: Integer);
    procedure   Do_OnMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure   Changed;
  Public
    constructor Create( anOwner: TComponent);                          override;
    destructor  Destroy;                                               override;
    procedure   FillRandom;
    procedure   Mutate;
    procedure   MateWith( aWorm: TKnobsWorm);
    procedure   MorphTo ( aWorm: TKnobsWorm);
    procedure   CopyFrom( aGene: TKnobsGene);                          overload;
    procedure   CopyFrom( aWorm: TKnobsWorm);                          overload;
    procedure   TriggerChange;
  public
    property    Gene                   : TKnobsGene          read FGene;
    property    Data[ anIndex: Integer]: TKnobsWormElt       read GetData; default;
    property    OnWormChanged          : TKnobsOnWormChanged read FOnWormChanged        write FOnWormChanged;
  published
    property    Align;
    property    Size                   : Integer             read GetSize               write SetSize               default 120;
    property    Audible                : Boolean             read FAudible              write SetAudible            default False;
    property    Frozen                 : Boolean             read FFrozen               write SetFrozen             default False;
    property    Clearance              : Byte                read FClearance            write SetClearance          default 10;
    property    BackColor              : TColor              read FBackColor            write SetBackColor          default clWhite;
    property    BorderColorNormal      : TColor              read FBorderColorNormal    write SetBorderColorNormal  default clGray;
    property    BorderColorAudible     : TColor              read FBorderColorAudible   write SetBorderColorAudible default clYellow;
    property    BorderColorFrozen      : TColor              read FBorderColorFrozen    write SetBorderColorFrozen  default clBlue;
    property    BorderColorBoth        : TColor              read FBorderColorBoth      write SetBorderColorBoth    default clLime;
    property    PenColor               : TColor              read FPenColor             write SetPenColor           default clBlack;
    property    CrossOverProbability   : TKnobsPercentage    read FCrossOverProbability write FCrossOverProbability default 33;
    property    MutationProbability    : TKnobsPercentage    read FMutationProbability  write FMutationProbability  default 50;
    property    MutationRange          : TKnobsWormElt       read FMutationRange        write FMutationRange        default  5;
    property    MorphPercentage        : TKnobsPercentage    read FMorphPercentage      write FMorphPercentage      default 20;
    property    DrawLines              : Boolean             read FDrawLines            write SetDrawLines          default True;
  published
    property    OnAudible              : TNotifyEvent        read FOnAudible            write FOnAudible;
    property    OnFrozen               : TNotifyEvent        read FOnFrozen             write FOnFrozen;
  end;

  TKnobsWormPanel = class( TCustomPanel)
  private
    FWorms                : array of TKnobsWorm;
    FRowCount             : Integer;
    FColCount             : Integer;
    FSpacing              : Byte;
    FClearance            : Byte;
    FWormWidth            : Byte;
    FWormHeight           : Byte;
    FWormSize             : Integer;
    FBackColor            : TColor;
    FBorderColorNormal    : TColor;
    FBorderColorAudible   : TColor;
    FBorderColorFrozen    : TColor;
    FBorderColorBoth      : TColor;
    FPenColor             : TColor;
    FCrossOverProbability : TKnobsPercentage;
    FMutationProbability  : TKnobsPercentage;
    FMutationRange        : TKnobsWormElt;
    FMorphPercentage      : TKnobsPercentage;
    FDrawLines            : Boolean;
    FAudibleWorm          : TKnobsWorm;
    FFrozenWorm           : TKnobsWorm;
    FAudible              : Integer;
    FFrozen               : Integer;
    FOnAudible            : TNotifyEvent;
    FOnFrozen             : TNotifyEvent;
    FOnWormChanged        : TKnobsOnWormChanged;
  private
    procedure   UpdateWorms;
    procedure   PlaceWorms;
    procedure   UpdateCount;
    procedure   AddWorm;
    procedure   DelWorm;
    procedure   DoAudible( aSender: Tobject);
    procedure   DoFrozen ( aSender: TObject);
    procedure   DoResize ( aSender: TObject);
    procedure   DoWormChanged( const aSender: TObject; const aGene: TKnobsGene);
    function    GetWorm( anIndex: Integer): TKnobsWorm;
    function    GetData( wIndex, dIndex: Integer): TKnobsWormElt;
    function    GetWormCount: Integer;
    procedure   SetRowCount            ( aValue : Integer    );
    procedure   SetColCount            ( aValue : Integer    );
    procedure   SetAudible             ( aValue : Integer    );
    procedure   SetFrozen              ( aValue : Integer    );
    procedure   SetSpacing             ( aValue : Byte       );
    procedure   SetClearance           ( aValue : Byte       );
    procedure   SetWormWidth           ( aValue : Byte       );
    procedure   SetWormHeight          ( aValue : Byte       );
    procedure   SetWormSize            ( aValue : Integer    );
    procedure   SetBackColor           ( aValue : TColor     );
    procedure   SetBorderColorNormal   ( aValue : TColor     );
    procedure   SetBorderColorAudible  ( aValue : TColor     );
    procedure   SetBorderColorFrozen   ( aValue : TColor     );
    procedure   SetBorderColorBoth     ( aValue : TColor     );
    procedure   SetPenColor            ( aValue : TColor     );
    procedure   SetCrossOverProbability( aValue : TKnobsPercentage);
    procedure   SetMutationProbability ( aValue : TKnobsPercentage);
    procedure   SetMutationRange       ( aValue : TKnobsWormElt   );
    procedure   SetMorphPercentage     ( aValue : TKnobsPercentage);
    procedure   SetDrawLines           ( aValue : Boolean    );
  private
    procedure   RegisterWorm( const aWorm: TKnobsWorm);
    procedure   UnregisterWorm( const aWorm: TKnobsWorm);
  protected
    procedure   Notification( aComponent: TComponent; anOperation: TOperation); override;
    procedure   WMEraseBackground( var aMsg: TWMEraseBkgnd);           message WM_ERASEBKGND;
  public
    constructor Create( anOwner: TComponent);                                   override;
    procedure   FillRandom;
    procedure   Mutate;
    procedure   Mate;
    procedure   Morph;
    procedure   TriggerChange;
    procedure   SelectLeftNeighbour;
    procedure   SelectRightNeighbour;
    procedure   SelectAboveNeighbour;
    procedure   SeleectBelowNeighbour;
    procedure   ActivateHPrev;
    procedure   ActivateHNext;
    procedure   ActivateVPrev;
    procedure   ActivateVNext;
  public
    property    Worm[ anIndex       : Integer] : TKnobsWorm    read GetWorm;  default;
    property    Data[ wIndex, dIndex: Integer] : TKnobsWormElt read GetData;
    property    AudibleWorm                    : TKnobsWorm    read FAudibleWorm;
    property    FrozenWorm                     : TKnobsWorm    read FFrozenWorm;
    property    WormCount                      : Integer       read GetWormCount;
  published
    property    Align;
    property    BorderStyle;
    property    BorderWidth;
    property    BevelInner     default bvLowered;
    property    BevelOuter     default bvRaised;
    property    BevelWidth;
    property    Caption;
    property    Color;
    property    Enabled;
    property    Height         default 167;
    property    ParentColor;
    property    ParentShowHint;
    property    PopupMenu;
    property    ShowHint;
    property    TabOrder;
    property    TabStop;
    property    Visible;
    property    Width          default 167;
    property    RowCount             : Integer             read FRowCount             write SetRowCount             default   2;
    property    ColCount             : Integer             read FColCount             write SetColCount             default   2;
    property    Audible              : Integer             read FAudible              write SetAudible              default  -1;
    property    Frozen               : Integer             read FFrozen               write SetFrozen               default  -1;
    property    Spacing              : Byte                read FSpacing              write SetSpacing              default   2;
    property    Clearance            : Byte                read FClearance            write SetClearance            default  10;
    property    WormWidth            : Byte                read FWormWidth            write SetWormWidth            default  80;
    property    WormHeight           : Byte                read FWormHeight           write SetWormHeight           default  80;
    property    WormSize             : Integer             read FWormSize             write SetWormSize             default 120;
    property    BackColor            : TColor              read FBackColor            write SetBackColor            default clWhite;
    property    BorderColorNormal    : TColor              read FBorderColorNormal    write SetBorderColorNormal    default clGray;
    property    BorderColorAudible   : TColor              read FBorderColorAudible   write SetBorderColorAudible   default clYellow;
    property    BorderColorFrozen    : TColor              read FBorderColorFrozen    write SetBorderColorFrozen    default clBlue;
    property    BorderColorBoth      : TColor              read FBorderColorBoth      write SetBorderColorBoth      default clLime;
    property    PenColor             : TColor              read FPenColor             write SetPenColor             default clBlack;
    property    CrossOverProbability : TKnobsPercentage    read FCrossOverProbability write SetCrossOverProbability default  33;
    property    MutationProbability  : TKnobsPercentage    read FMutationProbability  write SetMutationProbability  default  50;
    property    MutationRange        : TKnobsWormElt       read FMutationRange        write SetMutationRange        default   5;
    property    MorphPercentage      : TKnobsPercentage    read FMorphPercentage      write SetMorphPercentage      default  20;
    property    DrawLines            : Boolean             read FDrawLines            write SetDrawLines            default True;
  published
    property    OnAudible            : TNotifyEvent        read FOnAudible            write FOnAudible;
    property    OnFrozen             : TNotifyEvent        read FOnFrozen             write FOnFrozen;
    property    OnWormChanged        : TKnobsOnWormChanged read FOnWormChanged        write FOnWormChanged;
  end;

  TKnobsGene = class
  private
    FData  : TKnobsWormData;
  private
    function    GetSize: Integer;
    procedure   SetSize( aValue: Integer);
    function    GetData( anIndex: Integer): TKnobsWormElt;
    procedure   SetData( anIndex: Integer; aValue: TKnobsWormElt);
  public
    constructor Create( aWorm: TKnobsWorm = nil);
    destructor  Destroy;                                               override;
    procedure   Save( var aFile: TextFile);
  public
    property    Size : Integer                         read GetSize write SetSize;
    property    Data[ anIndex: Integer]: TKnobsWormElt read GetData write SetData; default;
  end;

  TKnobsGeneBox = class( TCustomListBox)
  private
    FOnGetGeneName : TKnobsOnGetGeneName;
    FDragged       : TKnobsGene;
  private
    procedure   Do_OnDragDrop ( Sender, Source: TObject; X, Y: Integer);
    procedure   Do_OnDragOver ( Sender, Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean);
    procedure   Do_OnEndDrag  ( Sender, Target: TObject; X, Y: Integer);
    procedure   Do_OnMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure   AddWorm( const aName: string; const aWorm: TKnobsWorm);
  public
    constructor Create( anOwner: TComponent);                          override;
    destructor  Destroy;                                               override;
    procedure   Load  ( const aFileName: string);
    procedure   Append( const aFileName: string);
    procedure   Save  ( const aFileName: string);
    procedure   Clear;                                                 override;
  published
    property    Align;
    property    Anchors;
    property    BiDiMode;
    property    BorderStyle;
    property    Color;
    property    Columns;
    property    Constraints;
    property    Ctl3D;
    property    DragCursor;
    property    DragKind;
    property    Enabled;
    property    ExtendedSelect;
    property    Font;
    property    ImeMode;
    property    ImeName;
    property    IntegralHeight;
    property    ItemHeight;
    property    ParentBiDiMode;
    property    ParentColor;
    property    ParentCtl3D;
    property    ParentFont;
    property    ParentShowHint;
    property    PopupMenu;
    property    ShowHint;
    property    Sorted;
    property    Style;
    property    TabOrder;
    property    TabWidth;
    property    Visible;
  published
    property    OnGetGeneName: TKnobsOnGetGeneName read FOnGetGeneName write FOnGetGeneName;
    property    OnClick;
    property    OnContextPopup;
    property    OnDblClick;
    property    OnEnter;
    property    OnExit;
    property    OnKeyDown;
    property    OnKeyPress;
    property    OnKeyUp;
    property    OnMeasureItem;
  end;

  TKnobsBoolArray = array[ 0 .. MaxInt div SizeOf( Boolean) - 1] of Boolean;
  PKnobsBoolArray = ^TKnobsBoolArray;

  TSubstitution = class
  private
    FRule   : TStringList;
    FActive : PKnobsBoolArray;
    FCount  : Integer;
  private
    procedure   CheckBounds( anIndex: Integer);
    function    GetActive( anIndex: Integer): Boolean;
    procedure   SetActive( anIndex: Integer; aValue: Boolean);
    function    GetRule  ( anIndex: Integer): string;
    procedure   SetRule  ( anIndex: Integer; const aValue: string);
    procedure   SetCount ( aValue : Integer);
  public
    constructor Create;
    destructor  Destroy;                                               override;
    procedure   CopyFrom( aSubstitution: TSubstitution);
    procedure   Load( const aFilename: string);
    procedure   Save( const aFilename: string);
  public
    property    Active[ anIndex: Integer]: Boolean read GetActive write SetActive;
    property    Rule  [ anIndex: Integer]: string  read GetRule   write SetRule;
    property    Count                    : Integer read FCount    write SetCount;
  end;



implementation





type

  TCharSet = set of AnsiChar;

  function ScanName( const S: string; var p: Integer; aSeparators: TCharSet): string;
  var
    i : Integer;
  begin
    Result := '';

    while (( p <= Length( S)) and CharInSet( S[ p], aSeparators))
    do Inc( p);

    i := p;

    while (( i <= Length( S)) and not CharInSet( S[ i], aSeparators))
    do Inc( i);

    Result := Copy( S, p, i - p);
    p := i;
  end;

  function ScanInt( const S: string; var p: Integer; aSeparators: TCharSet): Integer;
  var
    i : Integer;
    N : string;
  begin
    N := '';

    while (( p <= Length( S)) and CharInSet( S[ p], aSeparators))
    do Inc( p);

    i := p;

    while (( i <= Length( S)) and not CharInSet( S[ i], aSeparators))
    do Inc( i);

    N := Copy( S, p, i - p);
    p := i;
    Result := StrToIntDef( N, 0);
  end;

  function ScanBool( const S: string; var p: Integer; aSeparators: TCharSet): Boolean;
  var
    i : Integer;
    N : string;
  begin
    N := '';

    while (( p <= Length( S)) and CharInSet( S[ p], aSeparators))
    do Inc( p);

    i := p;

    while (( i <= Length( S)) and not CharInSet( S[ i], aSeparators))
    do Inc( i);

    N := Copy( S, p, i - p);
    p := i;
    Result := UpperCase( N) = 'TRUE';
  end;


{ ========
  TKnobsWorm = class( TGraphicControl)
  private
    FGene                 : TKnobsGene;
    FAudible              : Boolean;
    FFrozen               : Boolean;
    FOnAudible            : TNotifyEvent;
    FOnFrozen             : TNotifyEvent;
    FClearance            : Byte;
    FBackColor            : TColor;
    FBorderColorNormal    : TColor;
    FBorderColorAudible   : TColor;
    FBorderColorFrozen    : TColor;
    FBorderColorBoth      : TColor;
    FPenColor             : TColor;
    FCrossOverProbability : TKnobsPercentage;
    FMutationProbability  : TKnobsPercentage;
    FMutationRange        : TKnobsWormElt;
    FMorphPercentage      : TKnobsPercentage;
    FOnWormChanged        : TKnobsOnWormChanged;
    FDrawLines            : Boolean;
  public
    property    Gene                   : TKnobsGene          read FGene;
    property    Data[ anIndex: Integer]: TKnobsWormElt       read GetData; default;
    property    OnWormChanged          : TKnobsOnWormChanged read FOnWormChanged        write FOnWormChanged;
  published
    property    Align;
    property    Size                   : Integer             read GetSize               write SetSize               default 120;
    property    Audible                : Boolean             read FAudible              write SetAudible            default False;
    property    Frozen                 : Boolean             read FFrozen               write SetFrozen             default False;
    property    Clearance              : Byte                read FClearance            write SetClearance          default 10;
    property    BackColor              : TColor              read FBackColor            write SetBackColor          default clWhite;
    property    BorderColorNormal      : TColor              read FBorderColorNormal    write SetBorderColorNormal  default clGray;
    property    BorderColorAudible     : TColor              read FBorderColorAudible   write SetBorderColorAudible default clYellow;
    property    BorderColorFrozen      : TColor              read FBorderColorFrozen    write SetBorderColorFrozen  default clBlue;
    property    BorderColorBoth        : TColor              read FBorderColorBoth      write SetBorderColorBoth    default clLime;
    property    PenColor               : TColor              read FPenColor             write SetPenColor           default clBlack;
    property    CrossOverProbability   : TKnobsPercentage    read FCrossOverProbability write FCrossOverProbability default 33;
    property    MutationProbability    : TKnobsPercentage    read FMutationProbability  write FMutationProbability  default 50;
    property    MutationRange          : TKnobsWormElt       read FMutationRange        write FMutationRange        default  5;
    property    MorphPercentage        : TKnobsPercentage    read FMorphPercentage      write FMorphPercentage      default 20;
    property    DrawLines              : Boolean             read FDrawLines            write SetDrawLines          default True;
  published
    property    OnAudible              : TNotifyEvent        read FOnAudible            write FOnAudible;
    property    OnFrozen               : TNotifyEvent        read FOnFrozen             write FOnFrozen;
  private
 }

    procedure   TKnobsWorm.MutePeers;
    var
      i : Integer;
    begin
      if Assigned( Parent)
      then begin
        with Parent
        do begin
          for i := 0 to ControlCount - 1
          do begin
            if Controls[ i] is TKnobsWorm
            then TKnobsWorm( Controls[ i]).Audible := False;
          end;
        end;
      end;
    end;

    procedure   TKnobsWorm.UnActivatePeers;
    var
      i : Integer;
    begin
      if Assigned( Parent)
      then begin
        With Parent
        do begin
          for i := 0 to ControlCount - 1
          do begin
            if Controls[ i] is TKnobsWorm
            then TKnobsWorm( Controls[ i]).Frozen:= False;
          end;
        end;
      end;
    end;

    function    TKnobsWorm.GetSize: Integer;
    begin
      Result := FGene.Size;
    end;

    procedure   TKnobsWorm.SetSize ( aValue: Integer);
    begin
      FGene.Size := aValue;
      if not ( csDestroying In ComponentState)
      then FillRandom;
    end;

    function    TKnobsWorm.GetData( anIndex: Integer): TKnobsWormElt;
    begin
      Result := FGene[ anIndex];
    end;

    procedure   TKnobsWorm.SetAudible( aValue: Boolean);
    begin
      if FAudible <> aValue
      then begin
        if aValue
        then MutePeers;

        FAudible := aValue;

        if FAudible and Assigned( FOnAudible)
        then FOnAudible( Self);

        Invalidate;
      end;
    end;

    procedure   TKnobsWorm.SetFrozen( aValue: Boolean);
    begin
      if FFrozen <> aValue
      then Begin
        if aValue
        then UnActivatePeers;

        FFrozen := aValue;

        if FFrozen and Assigned( FOnFrozen)
        then FOnFrozen( Self);

        Invalidate;
      end;
    end;

    procedure   TKnobsWorm.SetClearance;
    begin
      if aValue <> FClearance
      then begin
        FClearance := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsWorm.SetBackColor( aValue: TColor);
    begin
      if aValue <> FBackColor
      then begin
        FBackColor := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsWorm.SetBorderColorNormal( aValue: TColor);
    begin
      if aValue <> FBorderColorNormal
      then begin
        FBorderColorNormal := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsWorm.SetBorderColorAudible( aValue: TColor);
    begin
      if aValue <> FBorderColorAudible
      then begin
        FBorderColorAudible := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsWorm.SetBorderColorFrozen( aValue: TColor);
    begin
      if aValue <> FBorderColorFrozen
      then begin
        FBorderColorFrozen := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsWorm.SetBorderColorBoth( aValue: TColor);
    begin
      if aValue <> FBorderColorBoth
      then begin
        FBorderColorBoth := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsWorm.SetPenColor( aValue: TColor);
    begin
      if aValue <> FPenColor
      then begin
        FPenColor := aValue;
        Invalidate;
      end;
    end;

    procedure   TKnobsWorm.SetDrawLines( aValue: Boolean);
    begin
      if aValue <> FDrawLines
      then begin
        FDrawLines := aValue;
        Invalidate;
      end;
    end;

//   Protected

    procedure   TKnobsWorm.Paint; // Override;
    type
      TP = record
        X, Y : Single;
      end;

    var
      Pts  : array of TP;
      maxx : Single;
      maxy : Single;
      minx : Single;
      miny : Single;

      function P( anX, anY: Single): TP;
      begin
        With Result
        do begin
          X := anX;
          y := anY;
        end;
      end;

      procedure ScalePts;
      var
        dx : Single;
        dy : Single;
        i  : Integer;
      begin
        dx := maxx - minx + 1;

        if dx < 1
        then dx := 1;

        dy := maxy - miny + 1;

        if dy < 1
        then dy := 1;

        for i := 0 to Size
        do begin
          Pts[ i].X := FClearance + ( Width  - FClearance) * (( Pts[ i].X - minx) / dx) - 1;
          Pts[ i].Y := FClearance + ( Height - FClearance) * (( Pts[ i].Y - miny) / dy) - 1;
        end;
      end;

      procedure CalcPts;
      var
        i  : Integer;
        nx : Single;
        ny : Single;
      begin
        Pts[ 0] := P( 0, 0);
        maxx    := 0.0;
        maxy    := 0.0;
        minx    := 0.0;
        miny    := 0.0;

        for i := 0 to Size - 1
        do begin
          nx := Pts[ i].X + Sin( 2 * Pi * FGene[ i] / 256);
          ny := Pts[ i].Y + Cos( 2 * Pi * FGene[ i] / 256);

          if nx > maxx
          then maxx := nx
          else if nx < minx
          then minx := nx;

          if ny > maxy
          then maxy := ny
          else if ny < miny
          then miny := ny;

          Pts[ i + 1] := P( nx, ny);
        end;
      end;

    var               // Sel   // Act
      Colors : array[ Boolean, Boolean] of TColor;
      i      : Integer;
    begin
      Colors[ false, false] := BorderColorNormal;
      Colors[ false, true ] := BorderColorFrozen;
      Colors[ true , false] := BorderColorAudible;
      Colors[ true , true ] := BorderColorBoth;

      SetLength( Pts, Size + 1);

      CalcPts;
      ScalePts;

      with Canvas
      do begin
        Brush.Color := BackColor;

        with Pen
        do Begin
          Color := Colors[ Audible, Frozen];
          Width := 2;
        end;

        Rectangle( 1, 1, Width - 1, Height - 1);

        with Pen
        do Begin
          Color := PenColor;
          Width := 1;
        end;

        if FDrawLines
        then Begin
          MoveTo( Round( Pts[ 0].X), Round( Pts[ 0].Y));

          for i := 1 to Size
          do LineTo( Round( Pts[ i].X), Round( Pts[ i].Y));
        end
        else Begin
          for i := 1 to Size - 1
          do begin
            Ellipse(
              Round( Pts[ i].X) - 2,
              Round( Pts[ i].Y) - 2,
              Round( Pts[ i].X) + 2,
              Round( Pts[ i].Y) + 2
            );
          end;
        end;
      end;
    end;

    procedure   TKnobsWorm.Do_OnDragDrop( Sender, Source: TObject; X, Y: Integer);
    begin
      Audible := True;

      if ( Source is TKnobsWorm) and ( Source <> Self)
      then Begin
        if HasParent and ( Parent is TKnobsWormPanel)
        then TKnobsWormPanel( Parent).Mate;
      end
      else if Source is TKnobsGeneBox
      then begin
        with Source as TKnobsGeneBox
        do CopyFrom( FDragged);
      end;
    end;

    procedure   TKnobsWorm.Do_OnDragOver( Sender, Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean);
    begin
      Accept :=
        (( Source is TKnobsWorm) and ( Source <> Self)) or
        ( Source is TKnobsGeneBox)
        ;
    end;

    procedure   TKnobsWorm.Do_OnEndDrag( Sender, Target: TObject; X, Y: Integer);
    begin
    end;

    procedure   TKnobsWorm.Do_OnMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    begin
      if Button = mbLeft
      then begin
        Audible := True;
        BeginDrag( False, -1);
      end
      else if Button = mbRight
      then Frozen := True;
    end;

    procedure   TKnobsWorm.Changed;
    var
      aGene : TKnobsGene;
    begin
      if Audible and Assigned( FOnWormChanged)
      then begin
        aGene := TKnobsGene.Create( Self);
        try
          FOnWormChanged( Self, aGene);
        finally
          aGene.DisposeOf;
        end;
      end;
      Invalidate;
    end;

//  Public

    constructor TKnobsWorm.Create( anOwner: TComponent); // Override;
    begin
      inherited;
      FGene                := TKnobsGene.Create;
      Size                 := 120;
      Audible              := False;
      Frozen               := False;
      Clearance            := 10;
      BackColor            := clWhite;
      BorderColorNormal    := clGray;
      BorderColorAudible   := clYellow;
      BorderColorFrozen    := clBlue;
      BorderColorBoth      := clLime;
      PenColor             := clBlack;
      CrossOverProbability := 33;
      MutationProbability  := 50;
      MutationRange        :=  5;
      MorphPercentage      := 20;
      Width                := 80;
      Height               := 80;
      OnDragDrop           := Do_OnDragDrop;
      OnDragOver           := Do_OnDragOver;
      OnEndDrag            := Do_OnEndDrag;
      OnMouseDown          := Do_OnMouseDown;
      DragMode             := dmManual;
      DrawLines            := True;
      FillRandom;
    end;

    destructor  TKnobsWorm.Destroy; // Override;
    begin
      Size := 0;
      FGene.DisposeOf;
      inherited;
    end;

    procedure   TKnobsWorm.FillRandom;
    var
      i : Integer;
    begin
      for i := 0 to Size - 1
      do FGene[ i] := Random( 256);

      Changed;
    end;

    procedure   TKnobsWorm.Mutate;
    var
      i : Integer;
    begin
      for i := 0 to Size - 1
      do begin
        if Random( 100) < MutationProbability
        then FGene[ i] := ( Integer( 256 + FGene[ i]) + MutationRange - Random( 2 * MutationRange + 1) ) mod 256;
      end;
      Changed;
    end;

    procedure   TKnobsWorm.MateWith( aWorm: TKnobsWorm);
    var
      i : Integer;
      S : Boolean;
    begin
      if Assigned( aWorm)
      then begin
        S := Random( 100) < CrossoverProbability;       // todo : should this not be inside the loop?

        for i := 0 to Min( Size, aWorm.Size) - 1
        do begin
          if S
          then FGene[ i] := aWorm.FGene[ i];
          S := S xor ( Random( 100) < CrossoverProbability);
        end;

        Mutate; // Performs Changed.
      End;
    end;

    procedure   TKnobsWorm.MorphTo( aWorm: TKnobsWorm);
    var
      i : Integer;
    begin
      if Assigned( aWorm)
      then Begin
        for i := 0 to Min( Size, aWorm.Size) - 1
        do begin
          FGene[ i] :=
            ((
              (100 - MorphPercentage) *       FGene[ i] +
                     MorphPercentage  * aWorm.FGene[ i]
            ) div 100 ) mod 256;
        end;

        Changed;
      end;
    end;

    procedure   TKnobsWorm.CopyFrom( aGene: TKnobsGene);
    var
      i : Integer;
    begin
      if Assigned( aGene)
      then begin
        Size := aGene.Size;

        for i := 0 to Size - 1
        do FGene[ i] := aGene[ i];
      end;
      Changed;
    end;

    procedure   TKnobsWorm.CopyFrom( aWorm: TKnobsWorm);
    begin
      if Assigned( aWorm)
      then CopyFrom( aWorm.Gene);
    end;

    procedure   TKnobsWorm.TriggerChange;
    begin
      Changed;
    end;


{ ========
  TWormPanel = class( TPanel)
  private
    FWorms                : array of TKnobsWorm;
    FRowCount             : Integer;
    FColCount             : Integer;
    FSpacing              : Byte;
    FClearance            : Byte;
    FWormWidth            : Byte;
    FWormHeight           : Byte;
    FWormSize             : Integer;
    FBackColor            : TColor;
    FBorderColorNormal    : TColor;
    FBorderColorAudible   : TColor;
    FBorderColorFrozen    : TColor;
    FBorderColorBoth      : TColor;
    FPenColor             : TColor;
    FCrossOverProbability : TKnobsPercentage;
    FMutationProbability  : TKnobsPercentage;
    FMutationRange        : TKnobsWormElt;
    FMorphPercentage      : TKnobsPercentage;
    FDrawLines            : Boolean;
    FAudibleWorm          : TKnobsWorm;
    FFrozenWorm           : TKnobsWorm;
    FAudible              : Integer;
    FFrozen               : Integer;
    FOnAudible            : TNotifyEvent;
    FOnFrozen             : TNotifyEvent;
    FOnWormChanged        : TKnobsOnWormChanged;
  public
    property    Worm[ anIndex       : Integer] : TKnobsWorm    read GetWorm;  default;
    property    Data[ wIndex, dIndex: Integer] : TKnobsWormElt read GetData;
    property    AudibleWorm                    : TKnobsWorm    read FAudibleWorm;
    property    FrozenWorm                     : TKnobsWorm    read FFrozenWorm;
    property    WormCount                      : Integer       read GetWormCount;
  published
    property    Align;
    property    BorderStyle;
    property    BorderWidth;
    property    BevelInner     default bvLowered;
    property    BevelOuter     default bvRaised;
    property    BevelWidth;
    property    Caption;
    property    Color;
    property    Enabled;
    property    Height         default 167;
    property    ParentColor;
    property    ParentShowHint;
    property    PopupMenu;
    property    ShowHint;
    property    TabOrder;
    property    TabStop;
    property    Visible;
    property    Width          default 167;
    property    RowCount             : Integer             read FRowCount             write SetRowCount             default   2;
    property    ColCount             : Integer             read FColCount             write SetColCount             default   2;
    property    Audible              : Integer             read FAudible              write SetAudible              default  -1;
    property    Frozen               : Integer             read FFrozen               write SetFrozen               default  -1;
    property    Spacing              : Byte                read FSpacing              write SetSpacing              default   2;
    property    Clearance            : Byte                read FClearance            write SetClearance            default  10;
    property    WormWidth            : Byte                read FWormWidth            write SetWormWidth            default  80;
    property    WormHeight           : Byte                read FWormHeight           write SetWormHeight           default  80;
    property    WormSize             : Integer             read FWormSize             write SetWormSize             default 120;
    property    BackColor            : TColor              read FBackColor            write SetBackColor            default clWhite;
    property    BorderColorNormal    : TColor              read FBorderColorNormal    write SetBorderColorNormal    default clGray;
    property    BorderColorAudible   : TColor              read FBorderColorAudible   write SetBorderColorAudible   default clYellow;
    property    BorderColorFrozen    : TColor              read FBorderColorFrozen    write SetBorderColorFrozen    default clBlue;
    property    BorderColorBoth      : TColor              read FBorderColorBoth      write SetBorderColorBoth      default clLime;
    property    PenColor             : TColor              read FPenColor             write SetPenColor             default clBlack;
    property    CrossOverProbability : TKnobsPercentage    read FCrossOverProbability write SetCrossOverProbability default  33;
    property    MutationProbability  : TKnobsPercentage    read FMutationProbability  write SetMutationProbability  default  50;
    property    MutationRange        : TKnobsWormElt       read FMutationRange        write SetMutationRange        default   5;
    property    MorphPercentage      : TKnobsPercentage    read FMorphPercentage      write SetMorphPercentage      default  20;
    property    DrawLines            : Boolean             read FDrawLines            write SetDrawLines            default True;
  published
    property    OnAudible            : TNotifyEvent        read FOnAudible            write FOnAudible;
    property    OnFrozen             : TNotifyEvent        read FOnFrozen             write FOnFrozen;
    property    OnWormChanged        : TKnobsOnWormChanged read FOnWormChanged        write FOnWormChanged;
  private
}

    procedure   TKnobsWormPanel.UpdateWorms;
    var
      i : Integer;
    begin
      for i := 0 to WormCount - 1
      do begin
        with Worm[ i]
        do begin
          Clearance            := Self.Clearance;
          BackColor            := Self.BackColor;
          BorderColorNormal    := Self.BorderColorNormal;
          BorderColorAudible   := Self.BorderColorAudible;
          BorderColorFrozen    := Self.BorderColorFrozen;
          BorderColorBoth      := Self.BorderColorBoth;
          PenColor             := Self.PenColor;
          CrossOverProbability := Self.CrossOverProbability;
          MutationProbability  := Self.MutationProbability;
          MutationRange        := Self.MutationRange;
          MorphPercentage      := Self.MorphPercentage;
          DrawLines            := Self.DrawLines;
          Width                := WormWidth;
          Height               := WormHeight;
        end;
      end;
    end;

    procedure   TKnobsWormPanel.PlaceWorms;
    var
      i      : Integer;
      CurCol : Integer;
      CurRow : Integer;
    begin
      Width  := ColCount * ( WormWidth  + Spacing) + Spacing;
      Height := RowCount * ( WormHeight + Spacing) + Spacing;
      CurCol := 0;
      CurRow := 0;

      for i := 0 to RowCount * ColCount - 1
      do begin
        with Worm[ i]
        do begin
          Left := Spacing + ( CurCol * ( Width  + Spacing));
          Top  := Spacing + ( CurRow * ( Height + Spacing));
          Inc( CurCol);
          if CurCol >= ColCount
          then begin
            CurCol := 0;
            Inc( CurRow);
          end;
        end;
      end;
    end;

    procedure   TKnobsWormPanel.AddWorm;
    begin
      TKnobsWorm.Create( Self).Parent := Self;
    end;

    procedure   TKnobsWormPanel.DelWorm;
    begin
      Worm[ WormCount - 1].DisposeOf;
    end;

    procedure   TKnobsWormPanel.UpdateCount;
    var
      aCount : Integer;
    begin
      aCount := FRowCount * FColCount;

      while WormCount < aCount
      do AddWorm;

      while WormCount > aCount
      do DelWorm;

      PlaceWorms;
    end;

    procedure   TKnobsWormPanel.DoAudible( aSender: Tobject);
    begin
      if aSender is TKnobsWorm
      then begin
        FAudibleWorm := TKnobsWorm( aSender);
        FAudible     := TKnobsWorm( aSender).Tag;

        if Assigned( FOnAudible)
        then FOnAudible( Self);

        TriggerChange;
      end;
    end;

    procedure   TKnobsWormPanel.DoFrozen( aSender: TObject);
    begin
      if aSender is TKnobsWorm
      then begin
        FFrozenWorm := TKnobsWorm( aSender);
        FFrozen     := TKnobsWorm( aSender).Tag;

        if Assigned( FOnFrozen)
        then FOnFrozen( Self);
      end;
    end;

    procedure   TKnobsWormPanel.DoResize( aSender: TObject);
    begin
      PlaceWorms;
    end;

    procedure   TKnobsWormPanel.DoWormChanged( const aSender: TObject; const aGene: TKnobsGene);
    begin
      if Assigned( FOnWormChanged)
      then FOnWormChanged( Self, aGene);
    end;

    function    TKnobsWormPanel.GetWorm( anIndex: Integer): TKnobsWorm;
    begin
      if ( anIndex >= 0) and ( anIndex< WormCount)
      then Result := TKnobsWorm( FWorms[ anIndex])
      else Result := nil;
    end;

    function    TKnobsWormPanel.GetData( wIndex, dIndex: Integer): TKnobsWormElt;
    begin
      if ( wIndex>= 0) and ( wIndex< WormCount)
      then Result := Worm[ wIndex][ dIndex]
      else Result := 0;
    end;

    function    TKnobsWormPanel.GetWormCount: Integer;
    begin
      Result := Length( FWorms);
    end;

    procedure   TKnobsWormPanel.SetRowCount( aValue : Integer);
    begin
      if aValue <> FRowCount
      then begin
        FRowCount := aValue;
        UpdateCount;
      end;
    end;

    procedure   TKnobsWormPanel.SetColCount( aValue : Integer);
    begin
      if aValue <> FColCount
      then begin
        FColCount := aValue;
        UpdateCount;
      end;
    end;

    procedure   TKnobsWormPanel.SetAudible( aValue : Integer);
    begin
      if ( aValue >= 0) and ( aValue < WormCount)
      then Worm[ aValue].Audible := True;
    end;

    procedure   TKnobsWormPanel.SetFrozen( aValue : Integer);
    begin
      if ( aValue >= 0) and ( aValue < WormCount)
      then Worm[ aValue].Frozen := True;
    end;

    procedure   TKnobsWormPanel.SetSpacing( aValue : Byte);
    begin
      if aValue <> FSpacing
      then begin
        FSpacing := aValue;
        PlaceWorms;
      end;
    end;

    procedure   TKnobsWormPanel.SetClearance( aValue : Byte);
    begin
      if aValue <> FClearance
      then begin
        FClearance := aValue;
        UpdateWorms;
        PlaceWorms;
      end;
    end;

    procedure   TKnobsWormPanel.SetWormWidth( aValue : Byte);
    begin
      if aValue <> FWormWidth
      then begin
        FWormWidth := aValue;
        UpdateWorms;
        PlaceWorms;
      end;
    end;

    procedure   TKnobsWormPanel.SetWormHeight( aValue : Byte);
    begin
      if aValue <> FWormHeight
      then begin
        FWormHeight := aValue;
        UpdateWorms;
        PlaceWorms;
      end;
    end;

    procedure   TKnobsWormPanel.SetWormSize( aValue : Integer);
    var
      i : Integer;
    begin
      if aValue <> FWormSize
      then begin
        FWormSize := aValue;

        for i := 0 to WormCount - 1
        do Worm[ i].Size := FWormSize;
      end;
    end;

    procedure   TKnobsWormPanel.SetBackColor( aValue : TColor);
    begin
      if aValue <> FBackColor
      then begin
        FBackColor := aValue;
        UpdateWorms;
      end;
    end;

    procedure   TKnobsWormPanel.SetBorderColorNormal( aValue : TColor);
    begin
      if aValue <> FBorderColorNormal
      then begin
        FBorderColorNormal := aValue;
        UpdateWorms;
      end;
    end;

    procedure   TKnobsWormPanel.SetBorderColorAudible( aValue : TColor);
    begin
      if aValue <> FBorderColorAudible
      then begin
        FBorderColorAudible := aValue;
        UpdateWorms;
      End;
    end;

    procedure   TKnobsWormPanel.SetBorderColorFrozen( aValue : TColor);
    begin
      If aValue <> FBorderColorFrozen
      then begin
        FBorderColorFrozen := aValue;
        UpdateWorms;
      end;
    end;

    procedure   TKnobsWormPanel.SetBorderColorBoth( aValue : TColor);
    begin
      if aValue <> FBorderColorBoth
      then begin
        FBorderColorBoth := aValue;
        UpdateWorms;
      end;
    end;

    procedure   TKnobsWormPanel.SetPenColor( aValue : TColor);
    begin
      if aValue <> FPenColor
      then begin
        FPenColor := aValue;
        UpdateWorms;
      end;
    end;

    procedure   TKnobsWormPanel.SetCrossOverProbability( aValue : TKnobsPercentage);
    begin
      if aValue <> FCrossOverProbability
      then begin
        FCrossOverProbability := aValue;
        UpdateWorms;
      end;
    end;

    procedure   TKnobsWormPanel.SetMutationProbability( aValue : TKnobsPercentage);
    begin
      if aValue <> FMutationProbability
      then begin
        FMutationProbability := aValue;
        UpdateWorms;
      end;
    end;

    procedure   TKnobsWormPanel.SetMutationRange( aValue : TKnobsWormElt);
    begin
      if aValue <> FMutationRange
      then begin
        FMutationRange := aValue;
        UpdateWorms;
      end;
    end;

    procedure   TKnobsWormPanel.SetMorphPercentage( aValue : TKnobsPercentage);
    begin
      if aValue <> FMorphPercentage
      then begin
        FMorphPercentage := aValue;
        UpdateWorms;
      end;
    end;

    procedure   TKnobsWormPanel.SetDrawLines( aValue : Boolean);
    begin
      if aValue <> FDrawLines
      then begin
        FDrawLines := aValue;
        UpdateWorms;
      end;
    end;

// Protected

//  private

    procedure   TKnobsWormPanel.RegisterWorm( const aWorm: TKnobsWorm);
    begin
      SetLength( FWorms, WormCount + 1);
      FWorms[ WormCount - 1] := aWorm;
    end;

    procedure   TKnobsWormPanel.UnregisterWorm( const aWorm: TKnobsWorm);
    var
      p : Integer;
      i : Integer;
    begin
      p := -1;
      for i := 0 to WormCount - 1
      do begin
        if FWorms[ i] = aWorm
        then begin
          p := i;
          Break;
        end;
      end;
      if p >= 0
      then begin
        for i := p to WormCount - 2
        do FWorms[ i] := FWorms[ i + 1];
        SetLength( FWorms, WormCount - 1);
      end;
    end;

    procedure   TKnobsWormPanel.Notification( aComponent: TComponent; anOperation: TOperation); // Override;
    var
      i : Integer;
    begin
      if aComponent is TKnobsWorm
      then begin
        if anOperation = opInsert
        then begin
          RegisterWorm( TKnobsWorm( aComponent));

          for i := 0 to WormCount - 1
          do Worm[ i].Tag := i;

          with TKnobsWorm( aComponent)
          do begin
            OnAudible     := DoAudible;
            OnFrozen      := DoFrozen;
            OnWormChanged := DoWormChanged;
          end;
        end
        else if anOperation = opRemove
        then begin
          with TKnobsWorm( aComponent)
          do begin
            OnAudible     := nil;
            OnFrozen      := nil;
            OnWormChanged := nil;
          end;

          UnregisterWorm( TKnobsWorm( aComponent));

          for i := 0 to WormCount - 1
          do Worm[ i].Tag := i;
        end;
      end;
      inherited;
    end;

    procedure   TKnobsWormPanel.WMEraseBackground( var aMsg: TWMEraseBkgnd); // message WM_ERASEBKGND;
    begin
      if aMsg.DC <> 0
      then Canvas.Handle := aMsg.DC;

      try
        Canvas.Pen.Color   := Color;
        Canvas.Brush.Color := Color;
        Canvas.Rectangle( 0, 0, Width, Height);
      finally
        if aMsg.DC <> 0
        then Canvas.Handle := 0;
      end;

      aMsg.Result := 0;
    end;

//  Public

    constructor TKnobsWormPanel.Create( anOwner: TComponent); // Override;
    begin
      inherited;
      OnResize             := DoResize;
      DoubleBuffered       := True;
      BevelInner           := bvLowered;
      BevelOuter           := bvRaised;
      Caption              := '.';
      Height               := 167;
      Width                := 167;
      Spacing              :=   2;
      Clearance            :=  10;
      WormWidth            :=  80;
      WormHeight           :=  80;
      WormSize             := 120;
      BackColor            := clWhite;
      BorderColorNormal    := clGray;
      BorderColorAudible   := clYellow;
      BorderColorFrozen    := clBlue;
      BorderColorBoth      := clLime;
      PenColor             := clBlack;
      CrossOverProbability :=  33;
      MutationProbability  :=  50;
      MutationRange        :=   5;
      MorphPercentage      :=  20;
      DrawLines            := True;
      Audible              :=  -1;
      Frozen               :=  -1;
      RowCount             :=   2;
      ColCount             :=   2;
    end;

    procedure   TKnobsWormPanel.FillRandom;
    var
      i : Integer;
    begin
      for i := 0 to WormCount - 1
      do begin
        if i <> Frozen
        then Worm[ i].FillRandom;
      end;
    end;

    procedure   TKnobsWormPanel.Mutate;
    var
      i : Integer;
    begin
      for i := 0 to WormCount - 1
      do begin
        if i <> Frozen
        then Worm[ i].Mutate;
      end;
    end;

    procedure   TKnobsWormPanel.Mate;
    var
      i : Integer;
    begin
      for i := 0 to WormCount - 1
      do begin
        if ( i <> Audible) and ( i <> Frozen)
        then Worm[ i].CopyFrom( Worm[ Audible]);
      end;

      for i := 0 to WormCount - 1
      do begin
        if ( i <> Audible) and ( i <> Frozen)
        then Worm[ i].MateWith( Worm[ Frozen]);
      end;
    end;

    procedure   TKnobsWormPanel.Morph;
    begin
      AudibleWorm.MorphTo( FrozenWorm);
    end;

    procedure   TKnobsWormPanel.TriggerChange;
    begin
      AudibleWorm.TriggerChange;
    end;

    procedure   TKnobsWormPanel.SelectLeftNeighbour;
    begin
      Audible := ( Audible + WormCount - 1) mod WormCount;
    end;

    procedure   TKnobsWormPanel.SelectRightNeighbour;
    begin
      Audible := ( Audible + 1) mod WormCount;
    end;

    procedure   TKnobsWormPanel.SelectAboveNeighbour;
    var
      aNext : Integer;
    begin
      aNext := Audible - ColCount;

      if aNext = - ColCount
      then aNext := WormCount - 1
      else if aNext < 0
      then aNext := aNext + WormCount - 1;

      Audible := aNext;
    end;

    procedure   TKnobsWormPanel.SeleectBelowNeighbour;
    var
      aNext : Integer;
    begin
      aNext := Audible + ColCount;

      if aNext = WormCount + ColCount - 1
      then aNext := 0
      else if aNext >= WormCount
      then aNext := aNext - WormCount + 1;

      Audible := aNext;
    end;

    procedure   TKnobsWormPanel.ActivateHPrev;
    begin
      Frozen := ( Frozen + WormCount - 1) mod WormCount;
    end;

    procedure   TKnobsWormPanel.ActivateHNext;
    begin
      Frozen := ( Frozen + 1) mod WormCount;
    end;

    procedure   TKnobsWormPanel.ActivateVPrev;
    var
      aNext : Integer;
    begin
      aNext := Frozen - ColCount;

      if aNext = - ColCount
      then aNext := WormCount - 1
      else if aNext < 0
      then aNext := aNext + WormCount - 1;

      Frozen := aNext;
    end;

    procedure   TKnobsWormPanel.ActivateVNext;
    var
      aNext : Integer;
    begin
      aNext := Frozen + ColCount;

      if aNext = WormCount + ColCount - 1
      then aNext := 0
      else if aNext >= WormCount
      then aNext := aNext - WormCount + 1;

      Frozen := aNext;
    end;


{========
  TGene = class( TObject)
  private
    FData  : TKnobsWormData;
  public
    property    Size : Integer                         read GetSize write SetSize;
    property    Data[ anIndex: Integer]: TKnobsWormElt read GetData write SetData; default;
  private
}

    function    TKnobsGene.GetSize: Integer;
    begin
      Result := Length( FData)
    end;

    procedure   TKnobsGene.SetSize( aValue: Integer);
    begin
      if aValue <> Length( FData)
      then SetLength( FData, aValue);
    end;

    function    TKnobsGene.GetData( anIndex: Integer): TKnobsWormElt;
    begin
      Result := FData[ anIndex];
    end;

    procedure   TKnobsGene.SetData( anIndex: Integer; aValue: TKnobsWormElt);
    begin
      FData[ anIndex] := aValue;
    end;

//  Public

    constructor TKnobsGene.Create( aWorm: TKnobsWorm);
    begin
      inherited Create;

      if Assigned( aWorm)
      then begin
        Size := aWorm.Size;
        Move( aWorm.FGene.FData[ 0], FData[ 0], Size * SizeOf( TKnobsWormElt));
      end;
    end;

    destructor  TKnobsGene.Destroy;
    begin
      Size := 0;
      inherited;
    end;

    procedure   TKnobsGene.Save( var aFile: TextFile);
    var
      i : Integer;
    begin
      Write( aFile, Format( '%d', [ Size]));

      for i := 0 to Size - 1
      do Write( aFile, Format( ', %d', [ Data[ i]]));

      WriteLn( aFile, ';');
    end;



{ ========
  TGeneBox = class( TCustomListBox)
  private
    FOnGetGeneName : TKnobsOnGetGeneName;
    FDragged       : TKnobsGene;
  published
    property    Align;
    property    Anchors;
    property    BiDiMode;
    property    BorderStyle;
    property    Color;
    property    Columns;
    property    Constraints;
    property    Ctl3D;
    property    DragCursor;
    property    DragKind;
    property    Enabled;
    property    ExtendedSelect;
    property    Font;
    property    ImeMode;
    property    ImeName;
    property    IntegralHeight;
    property    ItemHeight;
    property    ParentBiDiMode;
    property    ParentColor;
    property    ParentCtl3D;
    property    ParentFont;
    property    ParentShowHint;
    property    PopupMenu;
    property    ShowHint;
    property    Sorted;
    property    Style;
    property    TabOrder;
    property    TabWidth;
    property    Visible;
  published
    property    OnGetGeneName: TKnobsOnGetGeneName read FOnGetGeneName write FOnGetGeneName;
    property    OnClick;
    property    OnContextPopup;
    property    OnDblClick;
    property    OnEnter;
    property    OnExit;
    property    OnKeyDown;
    property    OnKeyPress;
    property    OnKeyUp;
    property    OnMeasureItem;
  private
}

    procedure   TKnobsGeneBox.Do_OnDragDrop( Sender, Source: TObject; X, Y: Integer);
    begin
      if Source is TKnobsWorm
      then AddWorm( '', TKnobsWorm( Source));
    end;

    procedure   TKnobsGeneBox.Do_OnDragOver( Sender, Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean);
    begin
      Accept := ( Source is TKnobsWorm) and ( Source <> Self);
    end;

    procedure   TKnobsGeneBox.Do_OnEndDrag( Sender, Target: TObject; X, Y: Integer);
    begin
    end;

    procedure   TKnobsGeneBox.Do_OnMouseDown( Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    var
      i : Integer;
    begin
      if Button = mbLeft
      then begin
        with Sender as TKnobsGeneBox
        do begin
          i := ItemAtPos(Point( X, Y), True);

          if i >= 0
          then begin
            FDragged := TKnobsGene( Items.Objects[ i]);
            BeginDrag( False);
          end;
        end;
      end;
    end;

    procedure   TKnobsGeneBox.AddWorm( const aName: string; const aWorm: TKnobsWorm);
    var
      aGene : TKnobsGene;
      S     : string;
    begin
      aGene := TKnobsGene.Create( aWorm);
      S     := aName;

      if S = ''
      then begin
        if Assigned( OnGetGeneName)
        then OnGetGeneName( Self, aGene, S);
      end;

      if S <> ''
      then Items.AddObject( S, aGene);
    end;

//  Public

    constructor TKnobsGeneBox.Create( anOwner: TComponent); // Override;
    begin
      inherited;
      OnDragDrop  := Do_OnDragDrop;
      OnDragOver  := Do_OnDragOver;
      OnEndDrag   := Do_OnEndDrag;
      OnMouseDown := Do_OnMouseDown;
    end;

    destructor  TKnobsGeneBox.Destroy; // Override;
    begin
      if HasParent
      then Clear;

      inherited;
    end;

    procedure   TKnobsGeneBox.Load( const aFileName: string);
    const
      Separators : TCharSet = [ ' ', ^I, ':', '=', ',', '.', ';'];
    var
      aFile : TextFile;
      S     : string;
      Count : Integer;
      aName : string;
      i     : Integer;
      j     : Integer;
      p     : Integer;
      GCount: Integer;
      aGene : TKnobsGene;
    begin
      AssignFile( aFile, aFileName);
      Reset( aFile);

      try
        ReadLn( aFile, S);
        p     := 1;
        Count := ScanInt( S, p, Separators);

        for i := 1 to Count
        do begin
          ReadLn( aFile, S);
          p     := 1;
          aName := ScanName( S, p, Separators);

          if aName = ''
          then raise EKnobsInvalidFile.Create( 'Not a Friemel ?');

          GCount := ScanInt( S, p, Separators);
          aGene  := TKnobsGene.Create;

          with aGene
          do begin
            Size := GCount;

            for j := 0 to GCount - 1
            do Data[ j] := ScanInt( S, p, Separators);
          end;

          Items.AddObject( aName, aGene);
        End;
      finally
        CloseFile( aFile);
      end;
    end;

    procedure   TKnobsGeneBox.Append( const aFileName: string);
    var
      aFile : TextFile;
      i     : Integer;
    begin
      AssignFile( aFile, aFileName);

      if FileExists( aFileName)
      then System.Append( aFile)
      else Rewrite( aFile);

      try
        with Items
        do begin
          WriteLn( aFile, Count);

          for i := 0 to Count - 1
          do begin
            Write( aFile, Format( '%s = ', [ Items[ i]]));
            TKnobsGene( Objects[ i]).Save( aFile);
          end;
        end;
      finally
        CloseFile( aFile);
      end;
    end;

    procedure   TKnobsGeneBox.Save( const aFileName: string);
    begin
      DeleteFile( aFileName);
      Append( aFileName);
    end;

    procedure   TKnobsGeneBox.Clear;
    begin
      with Items
      do begin
        while Count > 0
        do begin
          Objects[ Count - 1].DisposeOf;
          Objects[ Count - 1] := nil;
          Delete( Count - 1)
        end;
      end;
    end;


{ ========
  TSubstitution = class
  private
    FRule   : TStringList;
    FActive : PKnobsBoolArray;
    FCount  : Integer;
  public
    property    Active[ anIndex: Integer]: Boolean read GetActive write SetActive;
    property    Rule  [ anIndex: Integer]: string  read GetRule   write SetRule;
    property    Count                    : Integer read FCount    write SetCount;
  private
}

    procedure   TSubstitution.CheckBounds( anIndex: Integer);
    begin
      if ( anIndex < 0) or ( anIndex >= FCount)
      then
        raise
          EKnobsBounds.
            CreateFmt(
              'TSubstitution : Index ( %d) out of bounds [ %d, %d]',
              [
                anIndex,
                0,
                FCount - 1
              ]
            );
    end;

    function    TSubstitution.GetActive( anIndex: Integer): Boolean;
    begin
      CheckBounds( anIndex);
      Result := FActive^[ anIndex];
    end;

    procedure   TSubstitution.SetActive( anIndex: Integer; aValue: Boolean);
    begin
      CheckBounds( anIndex);
      FActive^[ anIndex] := aValue;
    end;

    function    TSubstitution.GetRule( anIndex: Integer): string;
    begin
      CheckBounds( anIndex);

      if anIndex < FRule.Count
      then Result := FRule[ anIndex]
      else Result := '';
    end;

    procedure   TSubstitution.SetRule( anIndex: Integer; const aValue: string);
    begin
      CheckBounds( anIndex);

      while FRule.Count <= anIndex
      do FRule.Add( '');

      FRule[ anIndex] := aValue;
    end;

    procedure   TSubstitution.SetCount ( aValue: Integer);
    var
      aBoolArray: PKnobsBoolArray;
    begin
      if aValue <> FCount
      then begin
        aBoolArray := nil;

        if aValue > 0
        then begin
          GetMem( aBoolArray, aValue * SizeOf( Boolean));
          FillChar( aBoolArray^, aValue * SizeOf( Boolean), 0);
        end;

        if ( aValue > 0) and ( FCount > 0)
        then Move( FActive^, aBoolArray^, Min( aValue, FCount) * SizeOf( Boolean));

        if FCount > 0
        then FreeMem( FActive, FCount * SizeOf( Boolean));

        FCount  := aValue;
        FActive := aBoolArray;

        while FRule.Count < Count
        do FRule.Add( '');
      end;
    end;

//  Public

    constructor TSubstitution.Create;
    begin
      inherited;
      FRule := TStringList.Create;
      FRule.Sorted := False;
    end;

    destructor  TSubstitution.Destroy; // Override;
    begin
      Count := 0;
      FRule.DisposeOf;
      inherited;
    end;

    procedure   TSubstitution.CopyFrom( aSubstitution: TSubstitution);
    var
      i : Integer;
    begin
      if Assigned( aSubstitution)
      then begin
        Count := aSubstitution.Count;
        FRule.Clear;

        for i := 0 to Count - 1
        do Rule[ i] := aSubstitution.Rule[ i];

        if Count > 0
        then Move( aSubstitution.FActive^, FActive^, Count * SizeOf( Boolean));
      end
      else Count := 0;
    end;

    procedure   TSubstitution.Load( const aFilename: string);
    const
      Separators : TCharSet = [ ' ', ^I];
    var
      aFile  : TextFile;
      S      : string;
      p      : Integer;
      i      : Integer;
    begin
      AssignFile( aFile, aFilename);
      Reset( aFile);

      try
        ReadLn( aFile, S);
        p     := 1;
        Count := ScanInt( S, p, Separators);
        FRule.Clear;

        for i := 0 to Count - 1
        do begin
          ReadLn( aFile, S);
          FRule.Add( S);
        end;

        for i := 0 to Count - 1
        do begin
          ReadLn( aFile, S);
          p          := 1;
          Active[ i] := ScanBool( S, p, Separators);
        end;
      finally
        CloseFile( aFile);
      end;
    end;


    procedure   TSubstitution.Save( const aFilename: string);
    const
      Bools : array[ Boolean] of string = ( 'False', 'True');
    var
      aFile : TextFile;
      i     : Integer;
    begin
      AssignFile( aFile, aFilename);
      Rewrite( aFile);

      try
        WriteLn( aFile, Count);

        for i := 0 to Count - 1
        do WriteLn( aFile, Rule[ i]);

        for i:= 0 to Count - 1
        do WriteLn( aFile, Bools[ Active[ i]]);
      finally
        CloseFile( aFile);
      end;
    end;



initialization

  RegisterClasses(
    [
      TKnobsWorm,
      TKnobsWormPanel
    ]
  );

finalization

  UnregisterClasses(
    [
      TKnobsWorm,
      TKnobsWormPanel
    ]
  );


end.
