unit SVFUnit;

{

   COPYRIGHT 2016 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    Oogstplein 6
    7545 HP Enschede
    the Netherlands
    http://www.iaf.nl/Users/BlueHell/
    http://bluehell.electro-music.com/
    j_dot.punter2@t2iaf_dot.nl

  All rights attributed to Blue Hell are owned by Jan Punter.


  original code :

    Original source : http://www.musicdsp.org/showone.php?id=142

    no copyright notice was found for that code.
}

interface

uses

  KnobsUtils, KnobsConversions, FormantControl;

type

  TSvfMode = (
    svfmLowPass,
    svfmBandPass,
    svfmBandReject,
    svfmHighPass
  );


  TSvfChamberlin = class
  private
    FInternalQ    : TSignal;
    FQ            : TSignal;
    FInternalFreq : TSignal;
    FFrequency    : TSignal;
    FSampleRate   : TSignal;
    FReciprocalSR : TSignal;
    FDelay1       : TSignal;
    FDelay2       : TSignal;
  private
    procedure   SetFrequency ( const aValue : TSignal);
    procedure   SetSampleRate( const aValue : TSignal);
    procedure   SetQ         ( const aValue : TSignal);
  public
    constructor Create;
    procedure   Process( const anInput : TSignal; var aLowPass, aBandPass, aBandReject, aHighPass : TSignal);
  public
    property    Frequency  : TSignal read FFrequency  write SetFrequency;
    property    SampleRate : TSignal read FSampleRate write SetSampleRate;
    property    Q          : TSignal read FQ          write SetQ;
  end;


  TSvfChamberlinChain = class
  private
    FFilters    : array of TSvfChamberlin;
    FLevel      : TSignal;
    FFrequency  : TSignal;
    FSampleRate : TSignal;
    FQ          : TSignal;
  private
    function    GetCount: Integer;
    procedure   SetFrequency ( const aValue : TSignal);
    procedure   SetSampleRate( const aValue : TSignal);
    procedure   SetQ         ( const aValue : TSignal);
  public
    constructor Create( aCount: Integer);
    destructor  Destroy;                                                                                       override;
    function    Process( const anInput : TSignal): TSignal;                                           virtual; abstract;
  public
    property    Count      : Integer read GetCount;
    property    Level      : TSignal read FLevel      write FLevel;
    property    Frequency  : TSignal read FFrequency  write SetFrequency;
    property    SampleRate : TSignal read FSampleRate write SetSampleRate;
    property    Q          : TSignal read FQ          write SetQ;
  end;


  TSvfChamberlinChainLowPass = class( TSvfChamberlinChain)
  public
    function    Process( const anInput : TSignal): TSignal;                                                    override;
  end;


  TSvfChamberlinChainBandPass = class( TSvfChamberlinChain)
  public
    function    Process( const anInput : TSignal): TSignal;                                                    override;
  end;


  TSvfChamberlinChainBandReject = class( TSvfChamberlinChain)
  public
    function    Process( const anInput : TSignal): TSignal;                                                    override;
  end;


  TSvfChamberlinChainHighPass = class( TSvfChamberlinChain)
  public
    function    Process( const anInput : TSignal): TSignal;                                                    override;
  end;


  TSvfChamberlinBank = class
  strict private
  const
    FILTER_COUNT = 17;
  private
    FLocked   : Boolean;
    FFilters  : array of TSvfChamberlinChain;
  private
    function    GetCount: Integer;
    procedure   SetLevel    ( const anIndex: Integer; const aValue: TSignal);
    procedure   SetFrequency( const anIndex: Integer; const aValue: TSignal);
    procedure   SetQ        ( const anIndex: Integer; const aValue: TSignal);
    procedure   SetLevels( const aValue: TSignal);
    procedure   SetQs( const aValue: TSignal);
  public
    constructor Create( aSampleRate: TSignal; aLength, aSteepness: Integer; aModeLow, aModeMid, aModeHigh: TSvfMode);
    destructor  Destroy;                                                                                       override;
    procedure   Process( const anInput : TSignal; out aValue : TSignal);
    procedure   ProcessSplit( const anInput : TSignal; var out1, out2, out3, out4, out5, out6, out7, out8, out9, out10, out11, out12, out13, out14, out15, out16, out17 : TSignal);
  public
    property    Count                              : Integer read GetCount;
    property    Level    [ const anIndex: Integer] : TSignal                write SetLevel;
    property    Frequency[ const anIndex: Integer] : TSignal                write SetFrequency;
    property    Q        [ const anIndex: Integer] : TSignal                write SetQ;
    property    Levels                             : TSignal                write SetLevels;
    property    Qs                                 : TSignal                write SetQs;
  end;


  TSvfFormantFilter = class
  private
    FFormantCollection : TFormantCollection;
    FLocked            : Boolean;
    FSampleRate        : TSignal;
    FFilters           : array of TSvfChamberlin;
    FLevels            : array of TSignal;
    FCurrentSet        : Integer;
    FFormantCount      : Integer;
    FVowelIndex        : TSignal;
    FVowel             : TVowel;
  private
    function    GetSetCount    : Integer;
    procedure   SetCurrentSet( anIndex: Integer);
    procedure   SetVowelIndex( anIndex: TSignal);
    procedure   SetSampleRate( aValue: TSignal);
  private
    procedure   RecreateFilter;
    procedure   UpdateFilter;
  public
    constructor Create( aFormantCollection: TFormantCollection; aSampleRate: TSignal);
    destructor  Destroy;                                                                                       override;
    procedure   Process( const anInput, aVowelIndex: TSignal; out aValue: TSignal);
  public
    property    SetCount   : Integer read GetSetCount;
    property    CurrentSet : Integer read FCurrentSet write SetCurrentSet;
    property    SampleRate : TSignal read FSampleRate write SetSampleRate;
  end;



implementation


uses

  sysutils;

const

  MIN_SAMPRATE = 44100.0;                 // Minimal sample rate, in Hz
  NORMAL_LIMIT = 1e-20;                   // Value below which the Normalize functions will truncate the value to zero {  -400 dB}


  function  Normalize( const aValue: TSignal): TSignal; inline;
  // This actually checks if aValue is zero, a denormal or it's absulote value being < 1e-20
  // and it treats all those cases as if they were zero.
  type
    TWordArray = array[ 0 .. 3] of Word;
    PWordArray = ^TWordArray;
  begin
    if (( PWordArray( @ aValue)^[ 3] and $7ff0) = 0) or ( Abs( aValue) < NORMAL_LIMIT)
    then Result := 0
    else Result := aValue;
  end;


{ ========
  TSvfChamberlin = class
  private
    FInternalQ    : TSignal;
    FQ            : TSignal;
    FInternalFreq : TSignal;
    FFrequency    : TSignal;
    FSampleRate   : TSignal;
    FReciprocalSR : TSignal;
    FDelay1       : TSignal;
    FDelay2       : TSignal;
  public
    property    Frequency  : TSignal read FFrequency  write SetFrequency;
    property    SampleRate : TSignal read FSampleRate write SetSampleRate;
    property    Q          : TSignal read FQ          write SetQ;
  private
}

    procedure   TSvfChamberlin.SetFrequency( const aValue : TSignal);
    var
      aFreq : TSignal;
    begin
      if aValue > FSampleRate / 6
      then aFreq := FSampleRate / 6
      else aFreq := aValue;

      if aFreq <> FFrequency
      then begin
        FFrequency    := aFreq;
        FInternalFreq := 2.0 * LookupSine( 0.5 * FFrequency * FReciprocalSR);
      end;
    end;


    procedure   TSvfChamberlin.SetSampleRate( const aValue : TSignal);
    var
      aRate : TSignal;
    begin
      if aValue <= MIN_SAMPRATE
      then aRate := MIN_SAMPRATE
      else aRate := aValue;

      if aRate <> FSampleRate
      then begin
        FSampleRate   := aValue;
        FReciprocalSR := 1 / FSampleRate;
        FInternalFreq := 2.0 * LookupSine( 0.5 * FFrequency * FReciprocalSR);
      end;
    end;


    procedure   TSvfChamberlin.SetQ( const aValue : TSignal);
    begin
      if aValue <> FQ
      then begin
        if aValue > 0.5
        then FQ := aValue
        else FQ := 0.5;

        FInternalQ := 1 / FQ;
      end;
    end;


//  public

    constructor TSvfChamberlin.Create;
    begin
      inherited;
      SampleRate := MIN_SAMPRATE;         // First set sample rate, then Frequency
      Frequency  := 329.6;                // in Hz
      Q          :=   1.0;                // Q can be set at arbitrary time
    end;


    procedure   TSvfChamberlin.Process( const anInput : TSignal; var aLowPass, aBandPass, aBandReject, aHighPass : TSignal);
    begin
      aLowPass    := Normalize( FDelay2 + FInternalFreq * FDelay1);
      aHighPass   := Normalize( anInput - aLowPass - FInternalQ * FDelay1);
      aBandPass   := Normalize( FInternalFreq * aHighPass + FDelay1);
      aBandReject := Normalize( aHighPass + aLowPass);
      FDelay1     := aBandPass;
      FDelay2     := aLowPass;
    end;


{ ========
  TSvfChamberlinChain = class
  private
    FFilters    : array of TSvfChamberlin;
    FLevel      : TSignal;
    FFrequency  : TSignal;
    FSampleRate : TSignal;
    FQ          : TSignal;
  public
    property    Count      : Integer read GetCount;
    property    Level      : TSignal read FLevel      write FLevel;
    property    Frequency  : TSignal read FFrequency  write SetFrequency;
    property    SampleRate : TSignal read FSampleRate write SetSampleRate;
    property    Q          : TSignal read FQ          write SetQ;
  private
}

    function    TSvfChamberlinChain.GetCount: Integer;
    begin
      Result := Length( FFilters);
    end;


    procedure   TSvfChamberlinChain.SetFrequency( const aValue : TSignal);
    var
      i : Integer;
    begin
      if aValue <> FFrequency
      then begin
        FFrequency := aValue;
        for i:= 0 to Count - 1
        do FFilters[ i].Frequency := FFrequency;
      end;
    end;


    procedure   TSvfChamberlinChain.SetSampleRate( const aValue : TSignal);
    var
      i : Integer;
    begin
      if aValue <> FSampleRate
      then begin
        FSampleRate := aValue;
        for i:= 0 to Count - 1
        do FFilters[ i].SampleRate := FSampleRate;
      end;
    end;


    procedure   TSvfChamberlinChain.SetQ( const aValue : TSignal);
    var
      i : Integer;
    begin
      if aValue <> FQ
      then begin
        FQ := aValue;
        for i:= 0 to Count - 1
        do FFilters[ i].Q := FQ / Count;
      end;
    end;


//  public

    constructor TSvfChamberlinChain.Create( aCount: Integer);
    var
      i : Integer;
    begin
      inherited Create;
      SetLength( FFilters, aCount);
      for i := 0 to Count - 1
      do FFilters[ i] := TSvfChamberlin.Create;
    end;


    destructor  TSvfChamberlinChain.Destroy; // override;
    var
      i : Integer;
    begin
      for i := 0 to Count - 1
      do FreeAndNil( FFilters[ i]);
      inherited;
    end;


{ ========
  TSvfChamberlinChainLowPass = class( TSvfChamberlinChain)
  public
}

    function    TSvfChamberlinChainLowPass.Process( const anInput : TSignal): TSignal; // override;
    var
      aBandPass   : TSignal;
      aBandReject : TSignal;
      aHighPass   : TSignal;
      i           : Integer;
    begin
      FFilters[ 0].Process( Normalize( FLevel * anInput), Result, aBandPass, aBandReject, aHighPass);
      for i := 1 to Count - 1
      do FFilters[ i].Process( Result, Result, aBandPass, aBandReject, aHighPass);
    end;


{ ========
  TSvfChamberlinChainBandPass = class( TSvfChamberlinChain)
  public
}

    function    TSvfChamberlinChainBandPass.Process( const anInput : TSignal): TSignal; // override;
    var
      aLowPass    : TSignal;
      aBandReject : TSignal;
      aHighPass   : TSignal;
      i           : Integer;
    begin
      FFilters[ 0].Process( Normalize( FLevel * anInput), aLowPass, Result, aBandReject, aHighPass);
      for i := 1 to Count - 1
      do FFilters[ i].Process( Result, aLowPass, Result, aBandReject, aHighPass);
    end;


{ ========
  TSvfChamberlinChainBandReject = class( TSvfChamberlinChain)
  public
}

    function    TSvfChamberlinChainBandReject.Process( const anInput : TSignal): TSignal; // override;
    var
      aLowPass  : TSignal;
      aBandPass : TSignal;
      aHighPass : TSignal;
      i         : Integer;
    begin
      FFilters[ 0].Process( Normalize( FLevel * anInput), aLowPass, aBandPass, Result, aHighPass);
      for i := 1 to Count - 1
      do FFilters[ i].Process( Result, aLowPass, aBandPass, Result, aHighPass);
    end;


{ ========
  TSvfChamberlinChainHighPass = class( TSvfChamberlinChain)
  public
}

    function    TSvfChamberlinChainHighPass.Process( const anInput : TSignal): TSignal; // override;
    var
      aLowPass    : TSignal;
      aBandPass   : TSignal;
      aBandReject : TSignal;
      i           : Integer;
    begin
      FFilters[ 0].Process( Normalize( FLevel * anInput), aLowPass, aBandPass, aBandReject, Result);
      for i := 1 to Count - 1
      do FFilters[ i].Process( Result, aLowPass, aBandPass, aBandReject, Result);
    end;


{ ========
  TSvfChamberlinBank = class
  private
    FLocked   : Boolean;
    FFilters  : array of TSvfChamberlinChain;
  public
    property    Count                              : Integer read GetCount;
    property    Level    [ const anIndex: Integer] : TSignal                write SetLevel;
    property    Frequency[ const anIndex: Integer] : TSignal                write SetFrequency;
    property    Q        [ const anIndex: Integer] : TSignal                write SetQ;
    property    Levels                             : TSignal                write SetLevels;
    property    Qs                                 : TSignal                write SetQs;
  private
}

    function    TSvfChamberlinBank.GetCount: Integer;
    begin
      Result := Length( FFilters);
    end;


    procedure   TSvfChamberlinBank.SetLevel( const anIndex: Integer; const aValue: TSignal);
    begin
      FFilters[ anIndex].Level := aValue;
    end;


    procedure   TSvfChamberlinBank.SetFrequency( const anIndex: Integer; const aValue: TSignal);
    begin
      FFilters[ anIndex].Frequency := aValue;
    end;


    procedure   TSvfChamberlinBank.SetQ( const anIndex: Integer; const aValue: TSignal);
    begin
      FFilters[ anIndex].Q := aValue;
    end;


    procedure   TSvfChamberlinBank.SetLevels( const aValue: TSignal);
    var
      i : Integer;
    begin
      for i := 0 to Count - 1
      do FFilters[ i].Level := aValue;
    end;


    procedure   TSvfChamberlinBank.SetQs( const aValue: TSignal);
    var
      i : Integer;
    begin
      for i := 0 to Count - 1
      do FFilters[ i].Q := aValue;
    end;


//  public

    constructor TSvfChamberlinBank.Create( aSampleRate: TSignal; aLength, aSteepness: Integer; aModeLow, aModeMid, aModeHigh: TSvfMode);

      function MakeFilter( aMode: TSvfMode; aSteepness: Integer): TSvfChamberlinChain;
      begin
        case aMode of
          svfmLowPass  : Result := TSvfChamberlinChainLowPass   .Create( aSteepness);
          svfmBandPass : Result := TSvfChamberlinChainBandPass  .Create( aSteepness);
          svfmHighPass : Result := TSvfChamberlinChainHighPass  .Create( aSteepness);
          else           Result := TSvfChamberlinChainBandReject.Create( aSteepness);
        end;
      end;

    var
      i : Integer;
    begin
      Assert( aSteepness > 0);
      FLocked := true;
      inherited Create;
      SetLength( FFilters, aLength);;

      for i := 0 to aLength - 1
      do begin
        if i = 0
        then FFilters[ i] := MakeFilter( aModeLow , aSteepness)
        else if i = aLength - 1
        then FFilters[ i] := MakeFilter( aModeHigh, aSteepness)
        else FFilters[ i] := MakeFilter( aModeMid , aSteepness);

        Level    [ i] := 0;
        Frequency[ i] := 329.6;
        FFilters[ i].SampleRate := aSampleRate;
      end;
      Qs := 0.5;

      FLocked := False;
    end;


    destructor  TSvfChamberlinBank.Destroy; // override;
    var
      i: Integer;
    begin
      FLocked := True;

      for i := 0 to Count - 1
      do FreeAndNil( FFilters[ i]);

      inherited;
    end;


    procedure   TSvfChamberlinBank.Process( const anInput : TSignal; out aValue : TSignal);
    var
      i : Integer;
    begin
      if FLocked
      then aValue := 0.0
      else begin
        aValue := 0.0;
        for i := 0 to Count - 1
        do aValue := Normalize( aValue + FFilters[ i].Process( anInput));
      end;
    end;


    procedure   TSvfChamberlinBank.ProcessSplit( const anInput : TSignal; var out1, out2, out3, out4, out5, out6, out7, out8, out9, out10, out11, out12, out13, out14, out15, out16, out17 : TSignal);
    begin
      if FLocked
      then begin
        out1  := 0.0;
        out2  := 0.0;
        out3  := 0.0;
        out4  := 0.0;
        out5  := 0.0;
        out7  := 0.0;
        out7  := 0.0;
        out8  := 0.0;
        out9  := 0.0;
        out10 := 0.0;
        out11 := 0.0;
        out12 := 0.0;
        out13 := 0.0;
        out14 := 0.0;
        out15 := 0.0;
        out16 := 0.0;
        out17 := 0.0;
      end
      else begin
        out1  := Normalize( FFilters[  0].Process( anInput));
        out2  := Normalize( FFilters[  1].Process( anInput));
        out3  := Normalize( FFilters[  2].Process( anInput));
        out4  := Normalize( FFilters[  3].Process( anInput));
        out5  := Normalize( FFilters[  4].Process( anInput));
        out7  := Normalize( FFilters[  5].Process( anInput));
        out7  := Normalize( FFilters[  6].Process( anInput));
        out8  := Normalize( FFilters[  7].Process( anInput));
        out9  := Normalize( FFilters[  8].Process( anInput));
        out10 := Normalize( FFilters[  9].Process( anInput));
        out11 := Normalize( FFilters[ 10].Process( anInput));
        out12 := Normalize( FFilters[ 11].Process( anInput));
        out13 := Normalize( FFilters[ 12].Process( anInput));
        out14 := Normalize( FFilters[ 13].Process( anInput));
        out15 := Normalize( FFilters[ 14].Process( anInput));
        out16 := Normalize( FFilters[ 15].Process( anInput));
        out17 := Normalize( FFilters[ 16].Process( anInput));
      end;
    end;


{ ========
  TSvfFormantFilter = class
  private
    FFormantCollection : TFormantCollection;
    FLocked            : Boolean;
    FSampleRate        : TSignal;
    FFilters           : array of TSvfChamberlin;
    FLevels            : array of TSignal;
    FCurrentSet        : Integer;
    FFormantCount      : Integer;
    FVowelIndex        : TSignal;
    FVowel             : TVowel;
  public
    property    SetCount   : Integer read GetSetCount;
    property    CurrentSet : Integer read FCurrentSet write SetCurrentSet;
    property    SampleRate : TSignal read FSampleRate write SetSampleRate;
  private
}

    function    TSvfFormantFilter.GetSetCount: Integer;
    begin
      if Assigned( FFormantCollection)
      then Result := FFormantCollection.SetCount
      else Result := 0;
    end;


    procedure   TSvfFormantFilter.SetCurrentSet( anIndex: Integer);
    begin
      if SetCount > 0
      then begin
        anIndex := Clip( anIndex, 0, SetCount - 1);

        if anIndex <> FCurrentSet
        then begin
          FCurrentSet := anIndex;
          RecreateFilter;
        end;
      end;
    end;


    procedure   TSvfFormantFilter.SetVowelIndex( anIndex: TSignal);
    begin
      if Assigned( FFormantCollection) and ( SetCount > 0)
      then begin
        if anIndex <> FVowelIndex
        then begin
          FVowelIndex := anIndex;
          FFormantCollection.Interpolate( FCurrentSet, FVowelIndex, FVowel);
          UpdateFilter;
        end;
      end;
    end;


    procedure   TSvfFormantFilter.SetSampleRate( aValue: TSignal);
    var
      i : Integer;
    begin
      if aValue <> FSampleRate
      then begin
        FSampleRate := aValue;

        for i := 0 to Length( FFilters) - 1
        do FFilters[ i].SampleRate := FSampleRate;
      end;
    end;


//  private

    procedure   TSvfFormantFilter.RecreateFilter;
    var
      i : Integer;
    begin
      FLocked := True;

      try
        for i := 0 to Length( FFilters) - 1
        do FreeAndNil( FFilters[ i]);

        if SetCount > 0
        then FFormantCount := FFormantCollection.FormantCount[ FCurrentSet, 0]
        else FFormantCount := 0;

        SetLength( FFilters       , FFormantCount);
        SetLength( FLevels        , FFormantCount);
        SetLength( FVowel.Formants, FFormantCount);

        for i := 0 to Length( FFilters) - 1
        do FFilters[ i] := TSvfChamberlin.Create;

        FVowelIndex := -1;
        SetVowelIndex( 0);
      finally
        FLocked := False;
      end;
    end;


    procedure   TSvfFormantFilter.UpdateFilter;
    var
      i : Integer;
    begin
      for i := 0 to FFormantCount - 1
      do begin
        FFilters[ i].Frequency := FVowel.Formants[ i].Freg;
        FFilters[ i].Q         := FVowel.Formants[ i].Q;
        FLevels [ i]           := FVowel.Formants[ i].LinLevel;
      end;
    end;


//  public

    constructor TSvfFormantFilter.Create( aFormantCollection: TFormantCollection; aSampleRate: TSignal);
    begin
      inherited Create;
      FLocked            := True;
      FVowel.Name1       := '';
      FVowel.Name2       := '';
      FFormantCount      :=  0;
      FCurrentSet        := -1;
      FVowelIndex        := -1;
      FFormantCollection := aFormantCollection;
      CurrentSet         := 0;                     // Will clear FLocked and will (re)create the filter
      SampleRate         := aSampleRate;
      SetVowelIndex( 0);                           // Will update the filter
    end;


    destructor  TSvfFormantFilter.Destroy; // override;
    var
      i : Integer;
    begin
      FLocked := True;

      for i := 0 to Length( FFilters) - 1
      do FreeAndNil( FFilters[ i]);

      inherited;
    end;


    procedure   TSvfFormantFilter.Process( const anInput, aVowelIndex: TSignal; out aValue: TSignal);
    var
      i           : Integer;
      aLowPass    : TSignal; // note : need to use distinct variables for all params
      aBandPass   : TSignal; //        passed to FFilters[ i].Process, or odd stuff
      aBandReject : TSignal; //        will happen (even when the values are not being
      aHighPass   : TSignal; //        used).
    begin
      aValue := 0.0;

      if not FLocked
      then begin
        SetVowelIndex( aVowelIndex);

        for i := 0 to Length( FFilters) - 1
        do begin
          FFilters[ i].Process( anInput, aLowPass, aBandPass, aBandReject, aHighPass);
          aValue := Normalize( aValue + FLevels[ i] * aBandPass);
        end;
      end;
    end;


end.

